/* Create trochoida */
tag_t create_trochoida(double tolerance, double start_point[3], 
					   double distances[3], double csys[9], 
					   double tool_diameter, double stock, 
					   double step, double start_angle, 
					   tag_t *edges_list, int count_edge, 
					   tag_t start_edge, double directions[3][3])
{
	tag_t spline_tag = NULL_TAG;
	int degree = 3, periodicity = 0, num_points, save_def_data = 1; 
	int irc, i, j, num_point = 1;
	double a, b, n, s, t = 0.0, l = 1.0;
	double trochoida_length;
	double curve_point[3], start_point_abs[3], to_x_axis[3], to_y_axis[3];
	double from_x_axis[3] = {1.0, 0.0, 0.0};
	double from_y_axis[3] = {0.0, 1.0, 0.0};
	double transform[16];
	double decimal, tmp_double, min_distance = -1.0, calc_dist;
	double pt_on_ent1[3], pt_on_ent2[3], guess1[3];
	double move_distance, move_vector[3], magnitude;
	double x_dir[3], ccw_dir[3], angle;
	double new_move_vector[3];
	char str[256] = "";
	
	UF_CURVE_pt_slope_crvatr_t *point_data = NULL;

	point_data = (void *)malloc(sizeof(point_data[0]));
	if(point_data == NULL) uc1601("Not memory for allocation", TRUE);
	a = distances[1]/2-stock-tool_diameter/2;
	trochoida_length = distances[0]-stock-tool_diameter/2;	
	decimal = modf((trochoida_length/(a*step)), &tmp_double);
	min_distance = tool_diameter/2+stock-stock/2;
	if(decimal <= 0.5)
	{
		if(trochoida_dir == 1) 
			n = floor(trochoida_length/(a*step))+0.25;
		else 
			n = floor(trochoida_length/(a*step))-0.25;
	}
	else
	{
		if(trochoida_dir == 1) 
			n = ceil(trochoida_length/(a*step))+0.25;
		else 
			n = ceil(trochoida_length/(a*step))-0.25;
	}

	b = (360*n)-a;
	if(trochoida_dir == 0) start_angle = start_angle - 180.0;

	num_points = (int)(250.0/step);

	start_point_abs[0] = 0.0;
	start_point_abs[1] = 0.0;
	start_point_abs[2] = 0.0;

	to_x_axis[0] = csys[0];
	to_x_axis[1] = csys[1];
	to_x_axis[2] = csys[2];
	to_y_axis[0] = csys[3];
	to_y_axis[1] = csys[4];
	to_y_axis[2] = csys[5];

	x_dir[0] = directions[0][0];
	x_dir[1] = directions[0][1];
	x_dir[2] = directions[0][2];
	ccw_dir[0] = directions[1][0];
	ccw_dir[1] = directions[1][1];
	ccw_dir[2] = directions[1][2];
	
	for(i = 0; i <= num_points; i++)
	{
		if(trochoida_dir == 1)
			s = ((1-t)*start_angle+t*b)/180.0*M_PI;
		else
			s = ((1-t)*start_angle-t*b)/180.0*M_PI;

		curve_point[0] = 
			start_point_abs[0]+(a*(t*n-l*sin(s))*step);
		curve_point[1] = 
			start_point_abs[1]+(a*(1-l*cos(s))-a);
		curve_point[2] = 
			start_point_abs[2];

		irc = UF_MTX4_csys_to_csys(start_point_abs, from_x_axis, from_y_axis,
					start_point, to_x_axis, to_y_axis, transform); 
		UF_MTX4_vec3_multiply(curve_point, transform, curve_point);
		
		for(j=0; j<count_edge; j++)
		{
			if(edges_list[j] != start_edge)
			{
				UF_MODL_ask_minimum_dist(edges_list[j],NULL_TAG, 0, 
					guess1, 1, curve_point, &calc_dist, 
					pt_on_ent1, pt_on_ent2);
				if(calc_dist < min_distance) 
				{
					move_distance = min_distance-calc_dist;
					move_vector[0] = curve_point[0]-pt_on_ent1[0];
					move_vector[1] = curve_point[1]-pt_on_ent1[1];
					move_vector[2] = curve_point[2]-pt_on_ent1[2];
					UF_VEC3_unitize(move_vector, tolerance, 
						&magnitude, new_move_vector);
					UF_VEC3_angle_between(new_move_vector, 
						x_dir, ccw_dir, &angle);
					if(angle <M_PI/2)
					{
						curve_point[0] = 
							curve_point[0]-(min_distance+calc_dist)*new_move_vector[0];
						curve_point[1] = 
							curve_point[1]-(min_distance+calc_dist)*new_move_vector[1];
						curve_point[2] = 
							curve_point[2]-(min_distance+calc_dist)*new_move_vector[2];
					}
					else 
					{
						curve_point[0] = 
							curve_point[0]+move_distance*new_move_vector[0];
						curve_point[1] = 
							curve_point[1]+move_distance*new_move_vector[1];
						curve_point[2] = 
							curve_point[2]+move_distance*new_move_vector[2];
					}
				}
			}
		}

		point_data[i].point[0] = curve_point[0];
		point_data[i].point[1] = curve_point[1];
		point_data[i].point[2] = curve_point[2];

		point_data[i].crvatr_type = UF_CURVE_CRVATR_NONE;
		point_data[i].crvatr[0] = 0.0;
		point_data[i].crvatr[1] = 0.0;
		point_data[i].crvatr[2] = 0.0;
		point_data[i].slope[0] = 0.0;
		point_data[i].slope[1] = 0.0;
		point_data[i].slope[2] = 0.0;

		if((i == 0) || (i >= num_points)) 
		{
			point_data[i].slope_type = UF_CURVE_SLOPE_AUTO;
		}
		else 
		{
			point_data[i].slope_type = UF_CURVE_SLOPE_NONE;
		}
		point_data = (void *)realloc(point_data, 
							sizeof(point_data[0])*(num_point+1));
		if(point_data == NULL) uc1601("Not memory for allocation !", TRUE);
		num_point++;
		t = t+(double)1/num_points;
	}
	irc = UF_CURVE_create_spline_thru_pts(degree, periodicity, 
		num_point-1, point_data, NULL, save_def_data, &spline_tag);
	if(irc) get_error(irc);

	return spline_tag;
}

extern void ufusr_cleanup(void){return;}

/* Find open edge of slot */
tag_t open_edge(tag_t *edges_array, int number_edges, double tolerance)
{
	int i_i, irc, is_equal;
	tag_t edge_1 = NULL_TAG;
	double junk[3], start[3], tangent [3], tangent_1[3];

	for(i_i=0; i_i < number_edges; i_i++)
	{
		irc = UF_MODL_ask_curve_props(edges_array[i_i], 0.0, start, 
			tangent, junk, junk, junk, junk);
		if(i_i == number_edges-1) i_i = -1;
		irc = UF_MODL_ask_curve_props(edges_array[i_i+1], 0.0, start, 
			tangent_1, junk, junk, junk, junk);
		if(irc) get_error(irc);
		UF_VEC3_is_equal(tangent, tangent_1, tolerance, &is_equal);
		if(is_equal != 1) 
		{
			edge_1 = edges_array[i_i];
			break;
		}
	}
	return edge_1;
}
/* Display temporary point */
static void display_temporary_point(double *coords)
{
    UF_OBJ_disp_props_t
        attrib = { 1, UF_OBJ_WHITE, UF_OBJ_NOT_BLANKED, UF_OBJ_WIDTH_NORMAL,
            UF_OBJ_FONT_SOLID, FALSE};

    UF_DISP_display_temporary_point(NULL_TAG, UF_DISP_USE_ACTIVE_PLUS,
        coords, &attrib, UF_DISP_ASTERISK);
}
/* Ask edge start point */
static void ask_edge_start(tag_t edge, double point[3])
{
	double limits[2];
    UF_EVAL_p_t eval;

	UF_EVAL_initialize(edge, &eval);
	UF_EVAL_ask_limits(eval, limits);
    UF_EVAL_evaluate(eval, 0, limits[0], point, NULL);
	UF_EVAL_free(eval);
}
/* Ask edge end point */
static void ask_edge_end(tag_t edge, double point[3])
{
	double limits[2];
    UF_EVAL_p_t eval;

	UF_EVAL_initialize(edge, &eval);
	UF_EVAL_ask_limits(eval, limits);
    UF_EVAL_evaluate(eval, 0, limits[1], point, NULL);
	UF_EVAL_free(eval);
}
/* Print string to listing windows */
void print_to_listing_windows(char str[256])
{
	char string[256];
	logical rsp;

	UF_UI_is_listing_window_open(&rsp);
	strcpy(string, str);
	if(rsp == FALSE) UF_UI_open_listing_window();
	UF_UI_write_listing_window(string);
	UF_UI_write_listing_window("\n");
	strcpy(string, "\0");
}
/* Unhighlight selected objects */
void unhighlight_objects(int count, tag_p_t object_p_t)
{
   int i;

   for(i = 0; i < count; i++)
   {
      UF_DISP_set_highlight(object_p_t[i], 0);
   }
}
/* Change tool parameters */
void change_tool_parameters(double tool_diameter, double tool_heigth,
		double tool_flute_ln, int tool_number,
		int tool_adj_reg, int tool_cutcom_reg, tag_t new_tool)
{
	int irc;

	irc = UF_PARAM_set_double_value(new_tool,
		UF_PARAM_TL_DIAMETER, tool_diameter);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_double_value(new_tool,
		UF_PARAM_TL_HEIGHT, tool_heigth);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_double_value(new_tool,
		UF_PARAM_TL_FLUTE_LN, tool_flute_ln);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(new_tool, 
		UF_PARAM_TL_NUMBER, tool_number);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(new_tool, 
		UF_PARAM_TL_ADJ_REG, tool_adj_reg);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(new_tool, 
		UF_PARAM_TL_CUTCOM_REG, tool_cutcom_reg);
	if(irc) get_error(irc);	
}
/* Change tool parameters for auto selection method */
void change_tool_parameters_auto(int tool_number, int tool_adj_reg, 
							int tool_cutcom_reg, tag_t new_tool)
{
	int irc;
	
	irc = UF_PARAM_set_int_value(new_tool, 
		UF_PARAM_TL_NUMBER, tool_number);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(new_tool, 
		UF_PARAM_TL_ADJ_REG, tool_adj_reg);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(new_tool, 
		UF_PARAM_TL_CUTCOM_REG, tool_cutcom_reg);
	if(irc) get_error(irc);	
}

/* Set operation parameters for finish operation */
static void set_operation_param_finish(tag_t operation, tag_t floor_face, 
									   tag_t from_tag, int from_gohome_flag)
{
	int irc;
	char str[256] = "";
	int value;
	tag_t ScaleTag, XDirTag, YDirTag, FloorTag, XAcisTag, YAcisTag, PointTags[2];
	
	depth_per_cut = 0.0;

	irc = UF_PARAM_set_int_value(operation, UF_PARAM_ENG_INIT_TYPE, 
		UF_PARAM_eng_init_type_automatic);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(operation, UF_PARAM_ENG_INTER_TYPE, 
		UF_PARAM_eng_inter_type_as_initial);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_double_value(operation,UF_PARAM_ENG_DIST, tool_diameter+0.1);
	if(irc) get_error(irc);

	irc = UF_PARAM_set_int_value(operation, UF_PARAM_RET_FINAL_TYPE, 
		UF_PARAM_ret_final_type_automatic);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(operation, UF_PARAM_RET_INTER_TYPE, 
		UF_PARAM_ret_inter_type_as_initial);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_double_value(operation,UF_PARAM_RET_AUTO_DIST, 0.0);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_double_value(operation,UF_PARAM_CLR_VERT, 0.0);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_double_value(operation,UF_PARAM_ENGRET_AUTO_OVERLAP, 0.0);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(operation, UF_PARAM_TRANSFER_METHOD, 
		UF_PARAM_transfer_method_previous_plane);
	irc = UF_PARAM_set_int_value(operation, UF_PARAM_ENGRET_AUTO_TYPE, 
		UF_PARAM_engret_auto_type_linear);

	irc = UF_PARAM_set_double_value(operation,
		UF_PARAM_CUTLEV_MAX_DEPTH, depth_per_cut);
	if(irc) get_error(irc);

	irc = UF_PARAM_set_int_value(operation, UF_PARAM_CUT_METHOD, 
		UF_PARAM_cut_method_standard_drive);
	if(irc) get_error(irc);
	irc = UF_PARAM_ask_int_value(operation, UF_PARAM_CUT_DIR_TYPE, 
		&value);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(operation, UF_PARAM_CUT_DIR_TYPE,
		UF_PARAM_cut_dir_conventional);
	if(irc) get_error(irc);

	irc = UF_SO_create_scalar_double(floor_face,UF_SO_update_after_modeling,
										1.0,&ScaleTag);
	if(irc) get_error(irc); 
	irc = UF_SO_create_scalar_double_dim(floor_face,UF_SO_update_after_modeling,
										  0.5,UF_SO_scalar_dimensionality_length,&XDirTag);
	if(irc) get_error(irc);
	irc = UF_SO_create_scalar_double_dim(floor_face,UF_SO_update_after_modeling,
										  0.5,UF_SO_scalar_dimensionality_length,&YDirTag);     
	if(irc) get_error(irc);
	irc = UF_SO_create_point_on_surface(floor_face,UF_SO_update_after_modeling,
										 floor_face,XDirTag,YDirTag,&PointTags[0]);
	if(irc) get_error(irc);
	irc = UF_SO_create_scalar_double_dim(floor_face,UF_SO_update_after_modeling,
										  0,UF_SO_scalar_dimensionality_length,&XDirTag);
	if(irc) get_error(irc);
	irc = UF_SO_create_scalar_double_dim(floor_face,UF_SO_update_after_modeling,
										  0,UF_SO_scalar_dimensionality_length,&YDirTag);     
	if(irc) get_error(irc);
	irc = UF_SO_create_point_on_surface(floor_face,UF_SO_update_after_modeling,
										 floor_face,XDirTag,YDirTag,&PointTags[1]);
	if(irc) get_error(irc);
	irc = UF_SO_create_dirr_two_points(floor_face,UF_SO_update_after_modeling,
										PointTags,&XAcisTag);
	if(irc) get_error(irc);
	irc = UF_SO_create_scalar_double_dim(floor_face,UF_SO_update_after_modeling,
										  0,UF_SO_scalar_dimensionality_length,&XDirTag);
	if(irc) get_error(irc);
	irc = UF_SO_create_scalar_double_dim (floor_face,UF_SO_update_after_modeling,
										  1,UF_SO_scalar_dimensionality_length,&YDirTag);     
	if(irc) get_error(irc);
	irc = UF_SO_create_point_on_surface (floor_face,UF_SO_update_after_modeling,
										 floor_face,XDirTag,YDirTag,&PointTags[1]);
	if(irc) get_error(irc);
	irc = UF_SO_create_dirr_two_points (floor_face,UF_SO_update_after_modeling,
										PointTags,&YAcisTag);
	if(irc) get_error(irc);
	irc = UF_SO_create_xform_pnt_xy_dirs (floor_face,UF_SO_update_after_modeling,
										  PointTags[0],XAcisTag,YAcisTag,ScaleTag,&FloorTag);
	irc = UF_PARAM_set_tag_value(operation,3007,FloorTag);
	if(irc) get_error(irc);

	if(from_gohome_flag == 1)
	{
		irc = UF_PARAM_set_int_value(operation, UF_PARAM_FROM_STATUS, 1);
		if(irc) get_error(irc);
		irc = UF_PARAM_set_tag_value(operation, UF_PARAM_FROM_POINT, from_tag);
		if(irc) get_error(irc);
	}
	else if(from_gohome_flag == 2)
	{
		irc = UF_PARAM_set_int_value(operation, UF_PARAM_GOHOME_STATUS, 1);
		if(irc) get_error(irc);
		irc = UF_PARAM_set_tag_value(operation, UF_PARAM_GOHOME_POINT, from_tag);
		if(irc) get_error(irc);
	}
}

/* Set operation parameters for rougth operation */
static void set_operation_param(tag_t operation, tag_t floor_face, tag_t from_tag, int from_gohome_flag)
{
	int irc;
	char str[256] = "";
	int value;
	tag_t ScaleTag, XDirTag, YDirTag, FloorTag, XAcisTag, YAcisTag, PointTags[2];

	irc = UF_PARAM_set_int_value(operation, UF_PARAM_ENG_INIT_TYPE, 
		UF_PARAM_eng_init_type_automatic);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(operation, UF_PARAM_ENG_INTER_TYPE, 
		UF_PARAM_eng_inter_type_as_initial);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_double_value(operation,UF_PARAM_ENG_DIST, tool_diameter+0.1);
	if(irc) get_error(irc);

	irc = UF_PARAM_set_int_value(operation, UF_PARAM_RET_FINAL_TYPE, 
		UF_PARAM_ret_final_type_automatic);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(operation, UF_PARAM_RET_INTER_TYPE, 
		UF_PARAM_ret_inter_type_as_initial);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_double_value(operation,UF_PARAM_RET_AUTO_DIST, 0.0);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_double_value(operation,UF_PARAM_CLR_VERT, 0.0);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_double_value(operation,UF_PARAM_ENGRET_AUTO_OVERLAP, 0.0);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(operation, UF_PARAM_TRANSFER_METHOD, 
		UF_PARAM_transfer_method_previous_plane);
	irc = UF_PARAM_set_int_value(operation, UF_PARAM_ENGRET_AUTO_TYPE, 
		UF_PARAM_engret_auto_type_linear);

	irc = UF_PARAM_set_double_value(operation,
		UF_PARAM_CUTLEV_MAX_DEPTH, depth_per_cut);
	if(irc) get_error(irc);

	irc = UF_PARAM_set_int_value(operation, UF_PARAM_CUT_METHOD, 
		UF_PARAM_cut_method_standard_drive);
	if(irc) get_error(irc);
	irc = UF_PARAM_ask_int_value(operation, UF_PARAM_CUT_DIR_TYPE, 
		&value);
	if(irc) get_error(irc);
	irc = UF_PARAM_set_int_value(operation, UF_PARAM_CUT_DIR_TYPE,
		UF_PARAM_cut_dir_conventional);
	if(irc) get_error(irc);

	irc = UF_SO_create_scalar_double (floor_face,UF_SO_update_after_modeling,
										1.0,&ScaleTag);
	if(irc) get_error(irc); 
	irc = UF_SO_create_scalar_double_dim (floor_face,UF_SO_update_after_modeling,
										  0.5,UF_SO_scalar_dimensionality_length,&XDirTag);
	if(irc) get_error(irc);
	irc = UF_SO_create_scalar_double_dim (floor_face,UF_SO_update_after_modeling,
										  0.5,UF_SO_scalar_dimensionality_length,&YDirTag);     
	if(irc) get_error(irc);
	irc = UF_SO_create_point_on_surface (floor_face,UF_SO_update_after_modeling,
										 floor_face,XDirTag,YDirTag,&PointTags[0]);
	if(irc) get_error(irc);
	irc = UF_SO_create_scalar_double_dim (floor_face,UF_SO_update_after_modeling,
										  0,UF_SO_scalar_dimensionality_length,&XDirTag);
	if(irc) get_error(irc);
	irc = UF_SO_create_scalar_double_dim (floor_face,UF_SO_update_after_modeling,
										  0,UF_SO_scalar_dimensionality_length,&YDirTag);     
	if(irc) get_error(irc);
	irc = UF_SO_create_point_on_surface (floor_face,UF_SO_update_after_modeling,
										 floor_face,XDirTag,YDirTag,&PointTags[1]);
	if(irc) get_error(irc);
	irc = UF_SO_create_dirr_two_points (floor_face,UF_SO_update_after_modeling,
										PointTags,&XAcisTag);
	if(irc) get_error(irc);
	irc = UF_SO_create_scalar_double_dim (floor_face,UF_SO_update_after_modeling,
										  0,UF_SO_scalar_dimensionality_length,&XDirTag);
	if(irc) get_error(irc);
	irc = UF_SO_create_scalar_double_dim (floor_face,UF_SO_update_after_modeling,
										  1,UF_SO_scalar_dimensionality_length,&YDirTag);     
	if(irc) get_error(irc);
	irc = UF_SO_create_point_on_surface (floor_face,UF_SO_update_after_modeling,
										 floor_face,XDirTag,YDirTag,&PointTags[1]);
	if(irc) get_error(irc);
	irc = UF_SO_create_dirr_two_points (floor_face,UF_SO_update_after_modeling,
										PointTags,&YAcisTag);
	if(irc) get_error(irc);
	irc = UF_SO_create_xform_pnt_xy_dirs (floor_face,UF_SO_update_after_modeling,
										  PointTags[0],XAcisTag,YAcisTag,ScaleTag,&FloorTag);
	irc = UF_PARAM_set_tag_value(operation,3007,FloorTag);
	if(irc) get_error(irc);
 
	if(from_gohome_flag == 1)
	{
		irc = UF_PARAM_set_int_value(operation, UF_PARAM_FROM_STATUS, 1);
		if(irc) get_error(irc);
		irc = UF_PARAM_set_tag_value(operation, UF_PARAM_FROM_POINT, from_tag);
		if(irc) get_error(irc);
	}
	else if(from_gohome_flag == 2)
	{
		irc = UF_PARAM_set_int_value(operation, UF_PARAM_GOHOME_STATUS, 1);
		if(irc) get_error(irc);
		irc = UF_PARAM_set_tag_value(operation, UF_PARAM_GOHOME_POINT, from_tag);
		if(irc) get_error(irc);
	}
}

void get_error(int irc)
{
	char err_msg[UF_UI_MAX_STRING_LEN+1];
	char msg[257];
	char fail_message[257];
	int  len;
	UF_get_fail_message(irc, fail_message);
	sprintf(err_msg, "error %d: %s", irc, fail_message);
	len = strlen(err_msg);
	len = (len < 255) ? len : 255;
	strncpy (msg, err_msg, len);
	msg[len+1] = '\000';
	uc1601(err_msg, TRUE);
	return;
}

int ufusr_ask_unload(void)
{
 return (UF_UNLOAD_IMMEDIATELY);
}

/* Selection initialization procedures. */
static int init_proc(UF_UI_selection_p_t select, void *user_data)
{
    int num_triples = 1;
    UF_UI_mask_t mask_triples[] = {UF_solid_type, 0, UF_UI_SEL_FEATURE_ANY_FACE};

/* Enable only fases */
    if((UF_UI_set_sel_mask(select,
                           UF_UI_SEL_MASK_CLEAR_AND_ENABLE_SPECIFIC,
                           num_triples, mask_triples)) == 0)
    {
        return (UF_UI_SEL_SUCCESS);
    }
    else
    {
        return (UF_UI_SEL_FAILURE);
    }
}

static void ask_curve_ends(tag_t curve, double ends[6])
{
    double junk[3];

    UF_MODL_ask_curve_props(curve, 0.0, &ends[0], junk, junk, junk,
        junk, junk);
    UF_MODL_ask_curve_props(curve, 1.0, &ends[3], junk, junk, junk,
        junk, junk);
}

static void sort_end_to_end(tag_t *curves, int n_curves)
{
	int ii, jj, kk, sofar = 1;
	double free_ends[6], gap, this_ends[6], tol;
	tag_t temp;

	UF_MODL_ask_distance_tolerance(&tol);
	ask_curve_ends(curves[0], free_ends);

	for(ii = 1; ii < n_curves; ii++)
	{
		ask_curve_ends(curves[ii], this_ends);
		for(jj = 0; jj < 4; jj = jj + 3)
		{
			UF_VEC3_distance(&this_ends[jj], &free_ends[0], &gap);
			if(gap <= tol)
			{
				UF_VEC3_copy(&this_ends[abs(jj-3)], &free_ends[0]);
				temp = curves[ii];
				for(kk = ii; kk > 0; kk--)
					curves[kk] = curves[kk-1];
				curves[0] = temp;
				ii = sofar;
				sofar++;
				break;
			}
			UF_VEC3_distance(&this_ends[jj], &free_ends[3], &gap);
			if (gap <= tol)
			{
				UF_VEC3_copy(&this_ends[abs(jj-3)], &free_ends[3]);
				temp = curves[sofar];
				curves[sofar] = curves[ii];
				curves[ii] = temp;
				ii = sofar;
				sofar++;
				break;
			}
		}
	}
}

int queryToolDiameter(double tool_diam, double cor_radius, 
					  double fl_length)
{
	tag_t methdGrp_t = NULL_TAG;
	UF_CAM_db_object_t db_obj;
	char cls_name[133]="END_MILL_NON_INDEXABLE";
	char str[256] = "";
	char query[133]="", **attvalues, **attnames, *atttype; 
	int matchCount=0, count=0, count_1, feedIdx=-1, speedIdx=-1;
	UF_RSET_t  rset;
	int i=0, j=0, k=0, irc;
	double tmp_diam, tmp_fl_len;

	UF_CAM_ask_cutter_db_object(&db_obj);
	sprintf(query,"[DB(Diameter)]<=[%.3lf]&&[DB(CorRadMill)]==[%.3lf]",
		tool_diam, cor_radius);
	UF_LIB_execute_query(db_obj,cls_name,query,&matchCount,&rset);
	for(j=0; j < matchCount; j++)
	{
		UF_LIB_ask_rset_map(rset,&count,
			(const char ***)&attnames,
			(const char **)&atttype);
		for(i=0; i<count-1; i++)
		{
			UF_LIB_ask_record_in_rset(rset,j,&count_1,
				(const char ***)&attvalues);
			if(strcmp(attnames[i],"Diameter")==0)
			{
				tmp_diam = atof(attvalues[i]);
				tmp_fl_len = atof(attvalues[3]);
				if(tmp_diam >= tool_diameter*trochoida_width_min)
				{
					returned_tool_list[k] = (char *)UF_allocate_memory
						(sizeof(char)*(strlen(attvalues[0])+1), &irc);
					if(irc) get_error(irc);
					material_description[k] = (char *)UF_allocate_memory
						(sizeof(char)*(strlen(attvalues[5])+1), &irc);
					if(irc) get_error(irc);
					strcpy(material_description[k], attvalues[5]);
					strcpy(returned_tool_list[k], attvalues[0]);
					tool_diam_array[k] = tmp_diam;
					tool_flute_ln_array[k] = tmp_fl_len;
					k++;
				}
			}
		}
		number_of_tool_ret = k;
		UF_LIB_free_rset_map(count,(const char **)attnames,
			(const char *)atttype);
		UF_free_string_array(count,(char **)attvalues);
	}
	return(TRUE);
}

double calculate_tool_diameter(double min_width, double stock, int cut_type)
{
	double t_diameter = 0.0;
/* cut_type = 0 - line, cut_type = 1 - trochoida */
	if(cut_type == 0)
	{
		t_diameter = floor(min_width-stock*2);
		return t_diameter;
	} else
	{
		t_diameter = floor((min_width-stock*2)*trochoida_width);
		return t_diameter;
	}
}

double ask_min_width(tag_t *face_objects, int face_count)
{
	double min_slot_width = 10000.0;
	char str[256] = "";

	int edge_count;
	int i, j, k, irc, edge_type, norm_dir, type;
	tag_t edge_object, max_edge, csys_id, matrix_id, start_edge;
	tag_t *list_edges;
	double arc_length, max_length = 0, tolerance, magnitude;
	double start_point[3], end_point[3], vector_dir[3], unit_vec1[3];
	double distances[3], tmp_p[3] = {0.0,0.0,0.0}, cross_product[3];
	double mtx[9], directions[3][3], dir[3];
	double min_dist, min_dist_1, pt_on_ent1[3], pt_on_ent2[3], guess1[3];
	double point[3], box[6], radius, rad_data;

	uf_list_p_t edge_list;

	UF_MODL_ask_distance_tolerance(&tolerance);
	for(i=0; i < face_count; i++)
	{
		irc = UF_MODL_ask_face_data(face_objects[i], &type, point, 
				dir, box, &radius, &rad_data, &norm_dir);
		if(irc) get_error(irc);

		irc = UF_MODL_ask_face_edges(face_objects[i], &edge_list);
		if(irc) get_error(irc);
		irc = UF_MODL_ask_list_count(edge_list, &edge_count);
		if(irc) get_error(irc);
		list_edges = (tag_t *)malloc(sizeof(tag_t));
		j = 1;
		for(k = 0; k < edge_count; k++)
		{
			irc = UF_MODL_ask_list_item(edge_list, k, &edge_object);
			if(irc) get_error(irc);
			list_edges[k] = edge_object;
			list_edges = (tag_t *)realloc(list_edges, sizeof(tag_t)*(j+1));
			j++;
			irc = UF_MODL_ask_edge_type(edge_object, &edge_type);
			if(irc) get_error(irc);
			if(edge_type == UF_MODL_LINEAR_EDGE)
			{
				irc = UF_CURVE_ask_arc_length(edge_object, 0.0, 1.0, 
					UF_MODL_UNITS_PART, &arc_length);
				if(irc) get_error(irc);
				if(arc_length > max_length) 
				{
					max_length = arc_length;
					max_edge = edge_object;
				}
			}
		}
		ask_edge_start(max_edge, start_point);
		ask_edge_end(max_edge, end_point);
		start_edge = open_edge(list_edges, edge_count, tolerance);

		UF_MODL_ask_minimum_dist(start_edge, NULL_TAG, 0, guess1, 
				1, start_point, &min_dist, pt_on_ent1, pt_on_ent2);
		UF_MODL_ask_minimum_dist(start_edge, NULL_TAG, 0, guess1, 
				1, end_point, &min_dist_1, pt_on_ent1, pt_on_ent2);

		if(min_dist > min_dist_1)
		{
			for(k = 0; k <= 2; k++)
			{
				tmp_p[k] = end_point[k];
				end_point[k] = start_point[k];
				start_point[k] = tmp_p[k];
			}
		}
		vector_dir[0] = end_point[0]-start_point[0];
		vector_dir[1] = end_point[1]-start_point[1];
		vector_dir[2] = end_point[2]-start_point[2];

		irc = UF_VEC3_unitize(vector_dir, tolerance, &magnitude, unit_vec1);
		if(irc) get_error(irc);

		UF_VEC3_negate(dir, dir);

		UF_VEC3_cross(vector_dir, dir, cross_product);

		irc = UF_MTX3_initialize(vector_dir, cross_product, mtx);
		if(irc) get_error(irc);

		irc = UF_CSYS_create_matrix(mtx, &matrix_id);
		if(irc) get_error(irc);
		irc = UF_CSYS_create_csys(start_point, matrix_id, &csys_id);
		if(irc) get_error(irc);

		irc = UF_MODL_ask_bounding_box_aligned(face_objects[i], csys_id, FALSE, 
				start_point, directions, distances);
		if(irc) get_error(irc);
		irc = UF_OBJ_delete_object(csys_id);
		if(irc) get_error(irc);	
		if(distances[1] <= min_slot_width)
			min_slot_width = distances[1];
	}
	UF_MODL_delete_list(&edge_list);
	free(list_edges);
	return min_slot_width;
}

static void traverseGroups(tag_t group)
{
	int ii=0, count=0;
	tag_t *members;
	logical is_group;
	char grp_name[133], oper_name[133];

	UF_NCGROUP_ask_member_list(group, &count, &members);

	for(ii = 0; ii < count; ii++)
	{
		UF_NCGROUP_is_group(members[ii], &is_group);
		if(is_group == TRUE)
		{
			UF_OBJ_ask_name(members[ii], grp_name);
			traverseGroups(members[ii]);
		}
		else
		{
			UF_OPER_ask_name_from_tag(members[ii], oper_name);
			operation_exist = 1;
		}
	}
	if(count > 0) UF_free(members);
}
